"""
Functions that help processing model data
"""
from struct import unpack
from pathlib import Path
from typing import Dict, Callable, Tuple, List, Union, IO, Optional

import numpy as np
from obspy.core import Trace, Stream, read as read_ms
from obspy.io.reftek.core import Reftek130

from sohstationviewer.model.downsampler import downsample
from sohstationviewer.model.mseed.blockettes_reader import (
    read_next_blkt, ReadBlocketteError,
)
from sohstationviewer.conf.dbSettings import dbConf
from sohstationviewer.conf import constants as const
from sohstationviewer.model.reftek.from_rt2ms import core
from sohstationviewer.view.util.enums import LogType


def read_soh_mseed(path2file: Path, file_name: str,
                   soh_streams: Dict[str, Dict[str, Stream]],
                   log_data: Dict[str, List[str]],
                   nets_in_file: Dict[Tuple[str, ...], str],
                   track_info: Callable[[str, LogType], None]
                   ) -> Dict[str, List]:
    """
    Use read() from obspy.core to read miniseed file:
       + if encoding is ASCII: log string will be added to log_data. The
           following blockettes' data if available will be converted to string
           to add to log_data as well.
       + otherwise traces from stream will be add to soh_streams to be
           merged later
    :param path2file: absolute path of mseed file
    :param file_name: name of mseed file
    :param soh_streams: holder of SOH mseed streams
    :param log_data: holder of info from log
    :param nets_in_file: holder of dict with key are
        nets of file and value is user-selected net so the rule will be
        applied for other files if their nets is subset of a key
    :param track_info: to display processing info
    """
    stream = read_ms(path2file)

    file = None
    nets = set()
    stats = set()
    net_stats = set()
    chan_ids = set()
    new_net_id = None
    for trace in stream:
        net_id = trace.stats['network'].strip()
        if net_id != "None":
            nets.add(net_id)
        if len(nets) > 1:
            for net in nets_in_file:
                net_set = set(net)
                if nets.issubset(net_set) or net_set.issubset(nets):
                    new_net_id = nets_in_file[net]
                    if nets != net_set:
                        nets_in_file[tuple(nets)] = new_net_id
                    break
            if new_net_id is None:
                new_net_id = list(nets)[0]
                nets_in_file[tuple(nets)] = new_net_id

    for trace in stream:
        if new_net_id is not None:
            # use one net ID for all traces in a file
            trace.stats['network'] = new_net_id
        net_id = trace.stats['network'].strip()
        chan_id = trace.stats['channel'].strip()
        sta_id = trace.stats['station'].strip()
        stats.add(sta_id)
        chan_ids.add(chan_id)
        net_stats.add((net_id, sta_id))
        if trace.stats.mseed['encoding'] == 'ASCII':
            file = read_ascii(
                path2file, file, sta_id, chan_id, trace, log_data, track_info)
        else:
            if sta_id not in soh_streams:
                soh_streams[sta_id] = {}
            if chan_id not in soh_streams[sta_id]:
                soh_streams[sta_id][chan_id] = Stream()
            soh_streams[sta_id][chan_id].append(trace)

    if file is not None:
        file.close()
    return {'nets': list(nets),
            'stats': sorted(list(stats)),
            'net_stats': sorted(list(net_stats)),
            'chan_ids': sorted(list(chan_ids))
            }


def read_soh_trace(trace: Trace) -> Dict:
    """
    Read SOH trace's info
    :param trace: mseed trace
    :return tr: with trace's info
        (structure in DataTypeModel.__init__.soh_ata[key][chan_id][orgTrace])
    """
    tr = {}
    tr['chanID'] = trace.stats['channel']
    tr['samplerate'] = trace.stats['sampling_rate']
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    tr['times'] = trace.times() + trace.stats['starttime'].timestamp
    tr['data'] = trace.data
    return tr


def read_mp_trace(trace: Trace) -> Dict:
    """
    Read mass possition trace's info using read_soh_trace(), then calculate
    real value for mass possition
    :param trace: mseed trace
    :return tr: with trace's info from read_soh_trace in which tr['data']
        has been converted from 16-bit signed integer in which
        32767= 2 ** 16/2 - 1 is the highest value of 16-bit two's complement
        number. The value is also multiplied by 10 for readable display.
        (structure in DataTypeModel.__init__.soh_data[key][chan_id][orgTrace])
    (According to 130_theory.pdf: Each channel connects to a 12-bit A/D
    converter with an input range of +/- 10V. These channel are read
    once per second as left-justified, 2's-compliment, 16 bit values.)

    """
    tr = read_soh_trace(trace)
    tr['data'] = np.round_(tr['data'] / 32767.0 * 10.0, 1)
    return tr


def read_waveform_trace(trace: Trace, sta_id: Union[Tuple[str, str], str],
                        chan_id: str, traces_info: List, tmp_dir: str) -> Dict:
    """
    read mseed waveform trace and save data to files to save mem for processing
        since waveform data are big.
    :param trace: mseed trace
    :param sta_id: station name
    :param chan_id: channel name
    :param traces_info: holder of traces_info, refer
        DataTypeModel.__init__.
        waveform_data[key]['read_data'][chan_id]['tracesInfo']
    :param tmp_dir: path to the directory that store memmap files
    :return tr: with trace's info
        (structure in DataTypeModel.__init__.
         waveform_data[key][chan_id][tracesInfo])
    """
    # gaps for SOH only for now
    tr = {}
    tr['samplerate'] = trace.stats['sampling_rate']
    tr['startTmEpoch'] = trace.stats['starttime'].timestamp
    tr['endTmEpoch'] = trace.stats['endtime'].timestamp
    """
    trace time start with 0 => need to add with epoch starttime
    times and data have type ndarray
    """
    times = trace.times() + trace.stats['starttime'].timestamp
    data = trace.data
    tr['size'] = times.size

    tr_idx = len(traces_info)
    tr['times_f'] = save_data_2_file(
        tmp_dir, 'times', sta_id, chan_id, times, tr_idx, tr['size'])
    tr['data_f'] = save_data_2_file(
        tmp_dir, 'data', sta_id, chan_id, data, tr_idx, tr['size'])
    return tr


def read_waveform_mseed(path2file: str, file_name: str,
                        sta_id: str, chan_id: str,
                        traces_info: List, data_time: List[float], tmp_dir: str
                        ) -> None:
    """
    Read traces from waveform mseed file to append to tracesInfo.
        data_time is update for new min and max time.
    :param path2file: absolute path to waveform mseed file
    :param file_name: name of waveform mseed file
    :param sta_id: station ID from indexing
    :param chan_id: channel ID from indexing
    :param traces_info: holder of traces_info, refer
        DataTypeModel.__init__.
        waveform_data[key]['read_data'][chan_id]['tracesInfo']
    :param data_time: DataTypeModel.__init__.data_time[key] - holder for data
        time of the current station
    :param tmp_dir: the folder to keep memmap files
    """
    stream = read_ms(path2file)
    for trace in stream:
        tr = read_waveform_trace(trace, sta_id, chan_id, traces_info, tmp_dir)
        data_time[0] = min(tr['startTmEpoch'], data_time[0])
        data_time[1] = max(tr['endTmEpoch'], data_time[1])
        traces_info.append(tr)


def read_waveform_reftek(rt130: Reftek130, key: Tuple[str, str],
                         read_data: Dict, data_time: List[float], tmp_dir: str
                         ) -> None:
    """
    Read traces from rt130 object to add to traces Info tracesInfo.
    dataTime is update for new min and max time.
    :param rt130: rt130 object
    :param key: (str, str) - key of the data set
    :param read_data: dict -holder of the readData refer
        DataTypeModel.__init__.waveform_data[key]['read_data']
    :param data_time: DataTypeModel.__init__.data_time[key] - holder for data
        time of the current station
    :param tmp_dir: str - the folder to keep memmap files
    """
    stream = core.Reftek130.to_stream(
        rt130,
        headonly=False,
        verbose=False,
        sort_permuted_package_sequence=True)
    for trace in stream:
        chan_id = trace.stats['channel']
        samplerate = trace.stats['sampling_rate']
        if chan_id not in read_data:
            read_data[chan_id] = {
                "tracesInfo": [],
                "samplerate": samplerate}
        traces_info = read_data[chan_id]['tracesInfo']
        tr = read_waveform_trace(trace, key, chan_id, traces_info, tmp_dir)
        data_time[0] = min(tr['startTmEpoch'], data_time[0])
        data_time[1] = max(tr['endTmEpoch'], data_time[1])
        traces_info.append(tr)


def read_ascii(path2file: Path, file: IO, sta_id: str, chan_id: str,
               trace: Trace, log_data: dict,
               track_info: Callable[[str, LogType], None]) -> IO:
    """
    Read mseed trace with ASCII encoding to add to log_data.
    :param path2file: str- absolute path of mseed file
    :param file: file object - to continue reading. Open new if file is None
    :param sta_id: str - station ID got from mseed header
    :param chan_id: str - channel ID got from mseed header
    :param trace: obspy.core.trace.Trace - mseed trace
    :param log_data: dict - holder for log messages, refer to
        DataTypeModel.__init__.log_data
    :param track_info: function - to display processing info
    :return file: file object - to close or continue to read
    """
    byteorder = trace.stats.mseed['byteorder']
    h = trace.stats
    log_text = "\n\nSTATE OF HEALTH: "
    log_text += ("From:%s  To:%s\n" % (h.starttime, h.endtime))
    text_from_data = trace.data.tobytes().decode()
    if text_from_data != '':
        log_text += text_from_data
    else:
        rec_len = h.mseed['record_length']
        if file is None:
            file = open(path2file, 'rb')
        data_bytes = file.read(rec_len)
        following_blkts_total = unpack('%s%s' % (byteorder, 'B'),
                                       data_bytes[39:40])[0]
        if following_blkts_total > 1:
            next_blkt_byte_no = 48 + 8  # header + blkt1000(SEED info)
        else:
            next_blkt_byte_no = 0
        while next_blkt_byte_no != 0:
            try:
                next_blkt_byte_no, info = read_next_blkt(
                    next_blkt_byte_no, data_bytes, byteorder)
                log_text += info
            except ReadBlocketteError as e:
                track_info(f"{sta_id} - {chan_id}: {e.msg}", LogType.ERROR)

    if sta_id not in log_data:
        log_data[sta_id] = {}
    if h.channel not in log_data[sta_id]:
        log_data[sta_id][h.channel] = []
    log_data[sta_id][h.channel].append(log_text)
    return file


def read_text(path2file: Path, file_name: str, text_logs: List) -> bool:
    """
    Read text file and add to log_data under channel TEXT. Raise exception
        if the file isn't a text file
    :param path2file: str - absolute path to text file
    :param file_name: str - name of text file
    :param text_logs: holder to keep log string, refer to
        DataTypeModel.__init__.log_data['TEXT']
    """
    with open(path2file, 'r') as file:
        try:
            content = file.read()
        except UnicodeDecodeError:
            raise Exception("Can't process file: %s" % file_name,
                            LogType.ERROR)

        log_text = "\n\n** STATE OF HEALTH: %s\n" % file_name
        log_text += content
        text_logs.append(log_text)

        return True


def save_data_2_file(tmp_dir: str, tm_data: str, sta_id: str, chan_id: str,
                     tr: np.ndarray, tr_idx: int, tr_size: int) -> Path:
    """
    Using np.memmap save time/data to file then free memory for processing
    :param tmp_dir: str - the temporary dir to save file in
    :param tm_data: str - "times"/"data"
    :param sta_id: str - station ID
    :param chan_id: str - channel ID
    :param tr: numpy array - of trace time or data
    :param tr_idx: int - trace index
    :param tr_size: int - trace size
    :return mem_file_name: str - name of memmap file
    """
    mem_file_name = Path(tmp_dir).joinpath(
        f"{sta_id}-{chan_id}-{tm_data}-{tr_idx}")
    mem_file = np.memmap(mem_file_name, dtype='int64', mode='w+',
                         shape=tr_size)
    mem_file[:] = tr[:]
    del tr
    del mem_file
    return mem_file_name


def check_chan(chan_id: str, req_soh_chans: List[str], req_wf_chans: List[str]
               ) -> Union[str, bool]:
    """
    Check if chanID is a requested channel.
    :param chan_id: str - channel ID
    :param req_soh_chans: list of str - requested SOH channels
    :param req_wf_chans: list of str - requested waveform channels
    :return: str/bool -
        'WF' if chanID is a requested waveform channel,
        'SOH' if chanID is a requested SOH channel,
        False otherwise.
    """
    ret = check_wf_chan(chan_id, req_wf_chans)
    if ret[0] == 'WF':
        if ret[1]:
            return "WF"
        else:
            return False
    if check_soh_chan(chan_id, req_soh_chans):
        return "SOH"
    return False


def check_soh_chan(chan_id: str, req_soh_chans: List[str]) -> bool:
    """
    Check if chan_id is a requested SOH channel.
    Mass position is always included.
    This function is used for mseed only so mass position is 'VM'.
    If there is no reqSOHChans, it means all SOH channels are requested
    :param chan_id: str - channel ID
    :param req_soh_chans: list of str - requested SOH channels
    :return: bool - True if chan_id is a requested SOH channel. False otherwise
    """
    if req_soh_chans == []:
        return True
    if chan_id in req_soh_chans:
        return True
    if 'EX?' in req_soh_chans and chan_id.startswith('EX'):
        if chan_id[2] in ['1', '2', '3']:
            return True
    if chan_id.startswith('VM'):
        if chan_id[2] in ['0', '1', '2', '3', '4', '5', '6']:
            return True
    return False


def check_wf_chan(chan_id: str, req_wf_chans: List[str]) -> Tuple[str, bool]:
    """
    Check if chanID is a waveform channel and is requested by user
    :param chan_id: str - channel ID
    :param req_wf_chans: list of str - requested waveform channels
    :return wf: str - '' if chan_id is not a waveform channel.
                      'WF' if chan_id is a waveform channel.
    :return has_chan: bool - True if chan_id is a requested waveform channel.
    """
    wf = ''
    has_chan = False
    if dbConf['seisRE'].match(chan_id):
        wf = 'WF'
    if req_wf_chans == ['*']:
        has_chan = True
    if chan_id in req_wf_chans:
        has_chan = True
    return wf, has_chan


def sort_data(data_dict: Dict) -> None:
    """
    Sort data in 'traces_info' in 'startTmEpoch' order
    :param data_dict: DataTypeModel.__init__.waveform_data
    """
    for sta_id in data_dict:
        for chan_id in data_dict[sta_id]['readData']:
            traces_info = data_dict[sta_id]['readData'][chan_id]['tracesInfo']
            traces_info = sorted(
                traces_info, key=lambda i: i['startTmEpoch'])


def squash_gaps(gaps: List[List[float]]) -> List[List[float]]:
    """
    Compress gaps from different channels that have time range related to
    each other to the ones with outside boundary (min start, max end).
    :param gaps: [[float, float],], [[float, float],] -
        list of gaps of multiple channels: [[start, end],], [[start, end],]
    :return: squashed_gaps: [[float, float],] - all related gaps are squashed
        extending to min start and max end [[min start, max end],]
    """
    squashed_gaps = []
    sgap_indexes = []
    for idx, g in enumerate(gaps):
        if idx in sgap_indexes:
            continue
        squashed_gaps.append(g)
        sgap_indexes.append(idx)
        for idx_, g_ in enumerate(gaps):
            if idx_ in sgap_indexes:
                continue
            if ((g[0] <= g_[0] <= g[1]) or (g[0] <= g_[1] <= g[1])
                    or (g_[0] <= g[0] <= g_[1]) or (g_[0] <= g[1] <= g_[1])):
                sgap_indexes.append(idx_)
                squashed_gaps[-1][0] = min(g[0], g_[0])
                squashed_gaps[-1][1] = max(g[1], g_[1])
            else:
                break
    return squashed_gaps


def trim_downsample_soh_chan(chan, start_tm, end_tm):
    """
    When channel is zoomed in, times and data will be trimmed off to exclude
        the non-included time. Then they will be downsampled to get to the
        expected const.CHAN_SIZE_LIMIT, and save to chan[times], chan[data]
        for plotting.

    :param chan: dict - channel info from soh_data, refer
        DataType_Model.__init_.soh_data[key][chan_id]
    :param start_tm: float - start time of zoomed section
    :param end_tm: float - end time of zoomed section
    """
    # zoom into the given time
    tr = chan['orgTrace']
    indexes = np.where((start_tm <= tr['times']) & (tr['times'] <= end_tm))
    if 'logIdx' in tr.keys():
        chan['times'], chan['data'], chan['logIdx'] = downsample(
            tr['times'][indexes], tr['data'][indexes], tr['logIdx'][indexes],
            rq_points=const.CHAN_SIZE_LIMIT)
    else:
        chan['times'], chan['data'], _ = downsample(
            tr['times'][indexes], tr['data'][indexes],
            rq_points=const.CHAN_SIZE_LIMIT)


def trim_waveform_data(wf_channel_data: Dict, start_time: float,
                       end_time: float) -> List[Dict]:
    """
    Trim off waveform traces whose times does not intersect the closed interval
    [start_time, end_time].

    :param wf_channel_data: dict - a dictionary that contains the waveform
        information of a channel. For its structure, refer to
        DataTypeModel.waveform_data[key]['readData'][chan_id].
    :param start_time: the start of the time range to trim waveform data to.
    :param end_time: the end of the time range to trim waveform data to.
    :return: a list with all the waveform traces that were not trimmed.
    """
    if ((start_time > wf_channel_data['tracesInfo'][-1]['endTmEpoch']) or
            (end_time < wf_channel_data['tracesInfo'][0]['startTmEpoch'])):
        return []

    good_start_indices = [index
                          for index, tr
                          in enumerate(wf_channel_data['tracesInfo'])
                          if tr['startTmEpoch'] > start_time]
    if good_start_indices:
        start_idx = good_start_indices[0]
        if start_idx > 0:
            start_idx -= 1  # start_time in middle of trace
    else:
        start_idx = 0

    good_end_indices = [idx
                        for idx, tr
                        in enumerate(wf_channel_data['tracesInfo'])
                        if tr['endTmEpoch'] <= end_time]
    if good_end_indices:
        end_idx = good_end_indices[-1]
        if end_idx < len(wf_channel_data['tracesInfo']) - 1:
            end_idx += 1  # end_time in middle of trace
    else:
        end_idx = 0
    end_idx += 1  # a[x:y+1] = [a[x], ...a[y]

    trimmed_traces_list = wf_channel_data['tracesInfo'][start_idx:end_idx]

    return trimmed_traces_list


def downsample_waveform_data(trimmed_traces_list: List[Dict],
                             start_time: float, end_time: float
                             ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    Downsample trimmed waveform data. Works on one trace at a time instead
    of combining the traces before downsampling.

    :param trimmed_traces_list: a list of traces that have been trimmed to
        include only those that intersect the closed interval
        [start_time, end_time]
    :param start_time: the start of the time range
    :param end_time: the end of the time range
    :return: list of downsampled times array and list of downsampled data array
    """
    # Calculate the number of requested_points
    total_size = sum([tr['size'] for tr in trimmed_traces_list])
    requested_points = 0
    if total_size > const.CHAN_SIZE_LIMIT:
        requested_points = int(
            const.CHAN_SIZE_LIMIT / len(trimmed_traces_list)
        )

    # Downsample the data
    downsampled_times_list = []
    downsampled_data_list = []
    for tr_idx, tr in enumerate(trimmed_traces_list):
        times = np.memmap(tr['times_f'],
                          dtype='int64', mode='r',
                          shape=tr['size'])
        data = np.memmap(tr['data_f'],
                         dtype='int64', mode='r',
                         shape=tr['size'])
        indexes = np.where((start_time <= times) & (times <= end_time))
        times = times[indexes]
        data = data[indexes]
        if requested_points != 0:
            times, data, _ = downsample(times, data,
                                        rq_points=requested_points)
        downsampled_times_list.append(times)
        downsampled_data_list.append(data)

    return downsampled_times_list, downsampled_data_list


def trim_downsample_wf_chan(chan: Dict, start_tm: float, end_tm: float,
                            firsttime: bool) -> Optional[bool]:
    """
    When channel is zoomed in, times and data will be trimmed off to exclude
        the non-included time.
    Different with soh_data where times and data are each in one np.array,
        in waveform_data, times and data are each kept in a list of np.memmap
        files along with startTmEpoch and endTmEpoch.
    Data in np.memmap files will be trimmed again if there is still time
        outside the zoom time range then downsampled and combined to get the
        times and data with size closed to the expected const.CHAN_SIZE_LIMIT,
        and save to chan[times], chan[data] for plotting.
    chan['fulldata']: is a flag set at the firsttime when channel isn't
        zoomed in and the data is small enough so it doesn't need to go through
        the process when zoomed in.

    :param chan: dict - channel info from soh_data, refer
        DataType_Model.__init_.soh_data[key][chan_id]
    :param start_tm: float - start time of zoomed section
    :param end_tm: float - end time of zoomed section
    :param firsttime: bool True for original size when channel is not zoomed in
    """
    if 'fulldata' in chan:
        # data is small, already has full in the first trim
        return

    z_traces_info = trim_waveform_data(chan, start_tm, end_tm)
    if not z_traces_info:
        return False

    total_size = sum([tr['size'] for tr in z_traces_info])
    if not firsttime and total_size > const.RECAL_SIZE_LIMIT:
        # size too big, recalc only when it's small enough
        return
    if total_size <= const.CHAN_SIZE_LIMIT and firsttime:
        chan['fulldata'] = True

    try:
        del chan['times']
        del chan['data']
    except Exception:
        pass

    downsampled_times, downsampled_data = downsample_waveform_data(
        z_traces_info, start_tm, end_tm
    )

    chan['times'] = np.hstack(downsampled_times)
    chan['data'] = np.hstack(downsampled_data)


def get_each_day_5_min_list(start_tm: float, end_tm: float) -> np.ndarray:
    """
    Get the list of all five minute for every day start from the day of startTm
        and end at the day of endTm.
    :param start_tm: float - start time
    :param end_tm: float - end time
    :return every_day_5_min_list: [[288 of floats], ] - the list of all start
        of five minutes for every day in which each day has 288 of 5 minutes.
    """
    exact_day_tm = (start_tm // const.SEC_DAY) * const.SEC_DAY
    exact_day_tm_list = []

    if start_tm < exact_day_tm:
        exact_day_tm_list = [exact_day_tm - const.SEC_DAY]

    while exact_day_tm < end_tm:
        exact_day_tm_list.append(exact_day_tm)
        exact_day_tm += const.SEC_DAY

    # list of start/end 5m in each day: every_day_5_min_list
    for idx, start_day_tm in enumerate(exact_day_tm_list):
        a_day_5_min = np.arange(start_day_tm,
                                start_day_tm + const.SEC_DAY,
                                const.SEC_5M)
        if idx == 0:
            every_day_5_min_list = np.array([a_day_5_min])
        else:
            every_day_5_min_list = np.vstack(
                (every_day_5_min_list, a_day_5_min))
    return every_day_5_min_list


def get_trim_tps_data(chan: Dict, start_tm: float, end_tm: float,
                      every_day_5_min_list: List[List[float]]
                      ) -> Optional[bool]:
    """
    Different with soh_data where times and data are each in one np.array,
        in waveform_data, times and data are each kept in a list of np.memmap
        files along with startTmEpoch and endTmEpoch.
    chan['startIdx'] and chan['endIdx'] will be identify to exclude np.memmap
        files that aren't in the zoom time range (startTm, endTm).
    Data in np.memmap will be trimmed according to times then time-power-square
        value for each 5 minutes will be calculated and saved in
        chan['tps-data']: np.mean(np.square(5m data))

    :param chan: dict - chan dictionary, refer to
        DataTypeModel.__init__.waveform_data[key]['readData'][chan_id]
    :param start_tm: float - start time of zoomed section
    :param end_tm: float - end time of zoomed section
    :param every_day_5_min_list: [[288 of floats], ] - the list of all start
        of five minutes for every day in which each day has 288 of 5 minutes.
    """
    # preset all 0 for all 5 minutes for each day
    tps_data = np.zeros((len(every_day_5_min_list), const.NO_5M_DAY))

    # zoom in to the given range
    chan['startIdx'] = 0
    chan['endIdx'] = len(chan['tracesInfo'])

    if ((start_tm > chan['tracesInfo'][-1]['endTmEpoch']) or
            (end_tm < chan['tracesInfo'][0]['startTmEpoch'])):
        return False

    indexes = [index for index, tr in enumerate(chan['tracesInfo'])
               if tr['startTmEpoch'] > start_tm]
    if indexes != []:
        chan['startIdx'] = indexes[0]
        if chan['startIdx'] > 0:
            chan['startIdx'] -= 1  # startTm in middle of trace
    else:
        chan['startIdx'] = 0

    indexes = [idx for (idx, tr) in enumerate(chan['tracesInfo'])
               if tr['endTmEpoch'] <= end_tm]
    if indexes != []:
        chan['endIdx'] = indexes[-1]
        if chan['endIdx'] < len(chan['tracesInfo']) - 1:
            chan['endIdx'] += 1  # endTm in middle of trace
    else:
        chan['endIdx'] = 0
    chan['endIdx'] += 1  # a[x:y+1] = [a[x], ...a[y]

    z_traces_info = chan['tracesInfo'][chan['startIdx']:chan['endIdx']]

    spr = chan['samplerate']
    chan['tps_data'] = []

    start_tps_tm = 0
    acc_data_list = []

    for tr_idx, tr in enumerate(z_traces_info):
        times = np.memmap(tr['times_f'],
                          dtype='int64', mode='r',
                          shape=tr['size'])
        data = np.memmap(tr['data_f'],
                         dtype='int64', mode='r',
                         shape=tr['size'])
        start_index = 0
        if tr_idx == 0:
            # get index of times with closet value to startTm
            start_index = np.abs(times - start_tm).argmin()
            start_tps_tm = times[start_index]

        # identify index in case of overlaps or gaps
        index = np.where((every_day_5_min_list <= times[start_index]) &
                         (every_day_5_min_list + const.SEC_5M >
                          times[start_index]))
        curr_row = index[0][0]
        curr_col = index[1][0]
        next_tps_tm = start_tps_tm + const.SEC_5M
        while end_tm >= next_tps_tm:
            next_index = int(start_index + spr * const.SEC_5M)
            if next_index >= tr['size']:
                acc_data_list.append(data[start_index:tr['size']])
                break
            else:
                acc_data_list.append(np.square(data[start_index:next_index]))
                acc_data = np.hstack(acc_data_list)
                if acc_data.size == 0:
                    tps_data[curr_row, curr_col] = 0
                else:
                    tps_data[curr_row, curr_col] = np.mean(acc_data)

                start_index = next_index
                curr_col += 1
                acc_data_list = []
                if curr_col == const.NO_5M_DAY:
                    curr_col = 0
                    curr_row += 1
                    next_tps_tm += const.SEC_5M
    chan['tps_data'] = tps_data


def find_tps_tm(given_tm: float, each_day_5_min_list: List[List[float]]
                ) -> Tuple[float, float]:
    """
    Find the position of the given time (given_tm) in time-power-squared plot
    :param given_tm: float - given time
    :param each_day_5_min_list: [[288 of floats], ] - the list of all start
        of five minutes for every day in which each day has 288 of 5 minutes.
    :return x_idx: int - index of time in the each_day_5_min_list
    :return y_idx: int - index of day plotted
                   (look in TimePowerSquaredWidget.getZoomData())
    """
    x_idx = None
    y_idx = None
    for day_idx, day in enumerate(each_day_5_min_list):
        for tm_idx, tm in enumerate(day):
            if tm > given_tm:
                y_idx = - day_idx
                x_idx = tm_idx - 1
                if x_idx < 0:
                    y_idx = -(day_idx - 1)
                    x_idx = len(day) - 1
                break
        if x_idx is not None:
            break
    return x_idx, y_idx


if __name__ == '__main__':
    print(check_chan('HH1', [], []))
