"""
MSeed object to hold and process MSeed data
"""

import os
from pathlib import Path
from typing import Dict, Tuple, List, Set

from obspy.core import Stream

from sohstationviewer.conf import constants
from sohstationviewer.controller.util import validate_file
from sohstationviewer.model.data_type_model import DataTypeModel, ThreadStopped
from sohstationviewer.model.handling_data import (
    read_waveform_mseed, squash_gaps, check_wf_chan, sort_data, read_soh_trace,
)
from sohstationviewer.model.mseed.from_mseedpeek.mseed_header import read_hdrs
from sohstationviewer.view.util.enums import LogType


class MSeed(DataTypeModel):
    """
    read and process mseed file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # channels: available channels got from files among
        # requested channels
        self.channels: Set[str] = set()

        # nets: available nets got from files
        self.nets: Set[str] = set()

        # stats: available stats got from files
        self.stats: Set[str] = set()

        """
        A file can have more than one experiment which only one of them should
        be correct as selected by user (handling_data.py - read_soh_mseed())
        nets_in_file is the dictionary with key is the tuple of all nets
        in a file, value is the selected net selected by user.
        """
        self.nets_in_file: Dict[Tuple[str, ...], str] = {}

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.read_soh_and_index_waveform(self.dir)

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.selected_key = self.select_sta_id()

        if self.selected_key is None:
            raise ThreadStopped()
        if len(self.req_wf_chans) != 0:
            self.read_wf_files(self.selected_key)

    def read_soh_and_index_waveform(self, folder: str):
        """
            + read waveform data for filename associate with time range
            + read soh_stream for unmerged traces
            + merge soh_stream into SOH data and mass position data
            + squash gaps from different soh channels to one
        :param folder: absolute path to data set folder
        """
        self.waveform_data, soh_stream = self.index_waveform(folder)
        self.soh_data, self.mass_pos_data, self.gaps = self.merge_soh_streams(
            soh_stream)

    def index_waveform(self, folder: str
                       ) -> Tuple[Dict, Dict[str, Dict[str, Stream]]]:
        """
        :param folder: absolute path to data set folder
        :return waveform_data: a dict for waveform data including
        {sta_id: {'filesInfo': {chan_id: [{filepath,time range,trace's info}]}
                 'readData': { chan_id: {samplerate
                         and preset traces_info which is a list of {info and
                         np.memmap's file path of trace data and time}
                  }
        }}
            (use np.memmap because data are too large to be merged and
                processed)
        :return soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        """
        soh_streams = {}
        waveform_data = {}
        net_stat_prop = {}
        stat_prop = {}
        chan_prop = {}
        count = 0

        for path, sub_dirs, files in os.walk(folder):
            for file_name in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()

                path2file = Path(path).joinpath(file_name)
                if not validate_file(path2file, file_name):
                    continue
                count += 1
                if count % 50 == 0:
                    self.track_info(
                        f'Read {count} file headers/ SOH files', LogType.INFO)

                ret = read_hdrs(
                    path2file, file_name, soh_streams, self.log_data,
                    self.req_soh_chans, self.req_wf_chans,
                    self.nets_in_file, self.track_info)
                if ret is None:
                    continue

                self.nets.update(ret['nets'])
                self.stats.add(ret['stats'][0])

                if len(ret['stats']) > 1:
                    k = tuple(ret['stats'])
                    if k not in stat_prop:
                        stat_prop[k] = []
                    stat_prop[k].append(file_name)
                else:
                    if len(ret['netStats']) > 1:
                        k = tuple(ret['netStats'])
                        if k not in net_stat_prop:
                            net_stat_prop[k] = []
                        net_stat_prop[k].append(file_name)
                if len(ret['chanIDs']) > 1:
                    k = tuple(ret['chanIDs'])
                    if k not in chan_prop:
                        chan_prop[k] = []
                    chan_prop[k].append(file_name)
                self.channels.update(ret['chanIDs'])
                if 'read' not in ret:
                    # not waveform
                    continue

                # TODO currently for waveform not read the file
                #  with 2 stations, 2 channels
                # correctly. Use the first one in the file only
                # When have time will dig more to this
                sta_id = ret['stats'][0]
                chan_id = ret['chanIDs'][0]
                if sta_id not in waveform_data:
                    waveform_data[sta_id] = {"filesInfo": {},
                                             "readData": {}}
                    self.data_time[sta_id] = [constants.HIGHEST_INT, 0]
                if chan_id not in waveform_data[sta_id]["filesInfo"]:
                    waveform_data[sta_id]["filesInfo"][chan_id] = []
                    waveform_data[sta_id]["readData"][chan_id] = {
                        "samplerate": ret['spr'],
                        "tracesInfo": []}
                waveform_data[sta_id]['filesInfo'][chan_id].append(ret)

        if len(stat_prop) > 0:
            errmsg = (f"More than one stations in a file: {stat_prop}. "
                      f"Will use the first one.")
            self.track_info(errmsg, LogType.ERROR)
        if len(net_stat_prop) > 0:
            errmsg = "More than one netIDs in a file: %s" % net_stat_prop
            self.track_info(errmsg, LogType.WARNING)
        if len(chan_prop) > 0:
            errmsg = (f"More than one channels in a file: {chan_prop} "
                      f"\nThis is a CRITICAL ERROR.")
            self.track_info(errmsg, LogType.ERROR)
        return waveform_data, soh_streams

    def merge_soh_streams(self, soh_streams: Dict[str, Dict[str, Stream]]
                          ) -> Tuple[dict, dict, dict]:
        """
        :param soh_streams: a dict of soh channel streams
            {sta_id: {chan_id: mseed stream of traces}}
        if there is more than one network for a station, user can either choose
        data for one network for that station, or combine all networks related
        to that station into one data set (by changing network name to the
        selected network then merge again)
        :return masspos_data: merged data for chan_id with 'VM' prefixed
            (structure in DataTypeModel.__init__.mass_pos_data)
        :return soh_data: merged data for other soh chan_ids
            (structure in DataTypeModel.__init__SOHData)
        :return gaps: start and end of gaps from all merged soh streams
            are squashed to the largest gaps
            (structure in DataTypeModel.__init__gaps)
        """

        soh_data = {}
        masspos_data = {}
        gaps = {}
        all_gaps: List[List[float]] = []
        for sta_id in soh_streams:
            self.stats.add(sta_id)
            self.data_time[sta_id] = [constants.HIGHEST_INT, 0]
            soh_data[sta_id] = {}
            masspos_data[sta_id] = {}
            for chan_id in soh_streams[sta_id]:
                self.channels.add(chan_id)
                stream = soh_streams[sta_id][chan_id]

                stream.merge()

                tr = stream[0]
                if len(stream) > 1:
                    nets = [tr.stats['network'].strip() for tr in stream]
                    nets += [f"Combine to {n}" for n in nets]
                    msg = (f"There are more than one net for sta {sta_id}.\n"
                           "Please select one or combine all to one.")
                    self.pause_signal.emit(msg, nets)
                    self.pause()
                    sel_net = nets[self.pause_response]
                    if "Combine" not in sel_net:
                        tr = [tr for tr in stream
                              if tr.stats['network'] == sel_net][0]
                        self.nets.add(sel_net)
                    else:
                        for tr in stream:
                            sel_net = sel_net.replace("Combine to ", "")
                            tr.stats['network'] = sel_net
                            self.nets.add(sel_net)
                        stream.merge()
                        tr = stream[0]

                gaps_in_stream = stream.get_gaps()
                all_gaps += [[g[4].timestamp, g[5].timestamp]
                             for g in gaps_in_stream]
                trace_info = read_soh_trace(tr)
                if chan_id.startswith('VM'):
                    masspos_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                else:
                    soh_data[sta_id][chan_id] = {
                        'orgTrace': trace_info}
                self.data_time[sta_id][0] = min(trace_info['startTmEpoch'],
                                                self.data_time[sta_id][0])
                self.data_time[sta_id][1] = max(trace_info['endTmEpoch'],
                                                self.data_time[sta_id][1])

            gaps[sta_id] = squash_gaps(all_gaps)
        return soh_data, masspos_data, gaps

    def select_sta_id(self) -> str:
        """
        :return selected_sta_id: the selected station id from self.stats
            + If there is only one station id, return it.
            + If there is more than one, show all ids, let user choose one to
                return.
        """

        stats = list(self.stats)
        selected_sta_id = stats[0]
        if len(stats) > 1:
            msg = ("There are more than one stations in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, stats)
            self.pause()
            selected_sta_id = stats[self.pause_response]

        self.track_info(f'Select Station {selected_sta_id}', LogType.INFO)
        return selected_sta_id

    def read_wf_files(self, sta_id: str) -> None:
        """
        From filesInfo, read all waveform data of requested waveform channels
        for given sta_id, in the selected time (from Main Window) to add
        to waveform_data[sat_id]['readData'][chan_id]['tracesInfo'].
        'tracesInfo' includes list of traces' info dict in startTmEpoch order.
        The keys of a trace's info dict:
         {
            'samplerate': sample rate of the data,
            'startTmEpoch': start epoch time of the trace,
            'endTmEpoch': end epoch time of the trace,
            'size': size of data,
            'times_f': np.memmap's file path for time
            'data_f': np.memmap's file path for data
        }
        :param sta_id: station ID
        """

        count = 0
        for chan_id in self.waveform_data[sta_id]['filesInfo']:
            # check chan_id
            has_chan = check_wf_chan(chan_id, self.req_wf_chans)
            if not has_chan:
                continue
            traces_info = self.waveform_data[sta_id][
                'readData'][chan_id]['tracesInfo']

            for file_info in self.waveform_data[sta_id]['filesInfo'][chan_id]:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped
                # file have been read
                if file_info['read']:
                    continue

                # check time
                has_data = False

                if ((self.read_start <= file_info['startEpoch'] <= self.read_end) or       # noqa: E501
                        (self.read_start <= file_info['endEpoch'] <= self.read_end)):      # noqa: E501
                    has_data = True
                if not has_data:
                    continue
                read_waveform_mseed(file_info['path2file'],
                                    file_info['fileName'],
                                    sta_id, chan_id, traces_info,
                                    self.data_time[sta_id], self.tmp_dir)
                file_info['read'] = True
                count += 1
                if count % 50 == 0:
                    self.track_info(
                        f'Read {count} waveform files', LogType.INFO)

        sort_data(self.waveform_data)
