from pathlib import Path
from typing import Dict, List, Tuple, Union
from sohstationviewer.view.util.enums import LogType
from sohstationviewer.conf import constants as const


def is_doc_file(file: Path, include_table_of_contents: bool = False)\
        -> bool:
    """
    Check if file is a document which is an '.help.md' file

    :param file: Absolute path to file
    :type file: Path
    :param include_table_of_contents: if Table of Contents file is considered
        as doc file or not
    :type include_table_of_contents: bool
    :return: True if file is a document, False otherwise
    :rtype: bool
    """
    if not file.is_file():
        return False
    if not file.name.endswith('.help.md'):
        return False
    if not include_table_of_contents and file.name == const.TABLE_CONTENTS:
        return False
    return True


def create_search_results_file(base_path: Path, search_text: str)\
        -> Path:
    """
    Create 'Search Results.md' file if not exist.
    Write file content which includes all links to '.help.md' files of which
        content contains search_text, excluding Table of Contents file.
    Format of each link is [name](URL)

    :param base_path: directory where document files are located
    :type base_path: Path
    :param search_text: text to search in each file
    :type search_text: str
    :return: path to search through file
    :rtype: Path
    """
    all_files = [f for f in list(base_path.iterdir()) if is_doc_file(f)]

    search_header = "# Search results\n\n"
    search_results = ""
    for file in sorted(all_files):
        with open(file) as f:
            content = f.read()
            if search_text in content:
                # space in URL must be replace with %20
                url_name = file.name.replace(" ", "%20")
                base_file_name = file.stem.split('.help')[0]
                try:
                    display_file_name = base_file_name.split(" _ ")[1]
                except IndexError:
                    display_file_name = base_file_name
                search_results += (f"+ [{display_file_name}]"
                                   f"({url_name})\n\n")
    if search_results == "":
        search_notfound = f"Text '{search_text}' not found."
        search_results = search_header + search_notfound
    else:
        search_found = (f"Text '{search_text}' found in the following files:"
                        f"\n\n---------------------------\n\n")
        search_results = search_header + search_found + search_results

    search_through_file = Path(base_path).joinpath(const.SEARCH_RESULTS)
    with open(search_through_file, "w") as f:
        f.write(search_results)
    return search_through_file


def create_table_of_content_file(base_path: Path) -> None:
    """
    Creating Table of Contents which includes all links to '.help.md' files.
    Format of each link is [name](URL)
    This function is added to __main__. So run functions.py to create "Table
        of Contents" file.

    :param base_path: directory where document files are located
    :type base_path: Path
    """
    all_files = [f for f in list(base_path.iterdir())
                 if is_doc_file(f, include_table_of_contents=True)]

    header = (
        "# SOH Station Viewer Documentation\n\n"
        "Welcome to the SOH Station Viewer documentation. Here you will find "
        "usage guides and other useful information in navigating and using "
        "this software.\n\n"
        "On the left-hand side you will find a list of currently available"
        " help topics.\n\n"
        "The home button can be used to return to this page at any time.\n\n"
        "# Table of Contents\n\n")
    links = ""

    for file in sorted(all_files):
        # space in URL must be replace with %20
        url_name = file.name.replace(" ", "%20")
        base_file_name = file.stem.split('.help')[0]
        try:
            display_file_name = base_file_name.split(" _ ")[1]
        except IndexError:
            display_file_name = base_file_name
        links += f"+ [{display_file_name}]({url_name})\n\n"

    contents = header + links

    contents_table_file = Path(base_path).joinpath(const.TABLE_CONTENTS)
    with open(contents_table_file, "w") as f:
        f.write(contents)
        print(f"{contents_table_file.absolute().as_posix()} has been created.")


def get_soh_messages_for_view(
        key: Union[str, Tuple[str, str]],
        soh_messages: Dict[str, Union[List[str], Dict[str, List[str]]]]) ->\
        Dict[str, List[str]]:
    """
    Convert SOH message of the selected key and TEXT log to dict of list of str
        to display

    :param key: key of selected data set: station_id
        or (experiment_number, serial)
    :type key: str/ (str, str)
    :param soh_messages: {'TEXT': log_text, key:{chan_id: sos_messages,},}
    :type soh_messages: Dict
    :return: {'TEXT'/chan_id: [str,]
    :rtype: Dict
    """

    soh_message_view = {}

    if 'TEXT' in soh_messages.keys() and soh_messages['TEXT'] != []:
        soh_message_view['TEXT'] = []
        for msg_list in soh_messages['TEXT']:
            for msg in msg_list.split('\n'):
                soh_message_view['TEXT'].append(msg)

    for chan_id in soh_messages[key]:
        soh_message_view[chan_id] = []

        for msg_lines in soh_messages[key][chan_id]:
            for msg in msg_lines.split('\n'):
                soh_message_view[chan_id].append(msg)
    return soh_message_view


def log_str(log_info: Tuple[str, LogType]) -> str:
    """
    Convert log_info to string that showing log line in saved file
    :param log_info: tuple include text and type of a log line
    :type log_info: Tuple[str, LogType]
    :return: line of log to save in file
    :rtype: str
    """
    log_text, log_type = log_info
    return f"{log_type.name}: {log_text}"


if __name__ == '__main__':
    create_table_of_content_file(Path('../../../documentation'))
